from enum import Enum
from types import NoneType
from typing import List, Type

import pytest

from jmux.awaitable import AwaitableValue, StreamableValues
from jmux.demux import JMux
from jmux.error import (
    EmptyKeyError,
    MissingAttributeError,
    ObjectAlreadyClosedError,
    ParsePrimitiveError,
    UnexpectedCharacterError,
)
from jmux.types import Mode, State

# fmt: off
parse_correct_stream__params = [
    ("", [], State.START),
    ("{", [Mode.ROOT], State.EXPECT_KEY),
    ("{ ", [Mode.ROOT], State.EXPECT_KEY),
    ('{"', [Mode.ROOT], State.PARSING_KEY),
    ('{"key_', [Mode.ROOT], State.PARSING_KEY),
    ('{"key_str', [Mode.ROOT], State.PARSING_KEY),
    ('{"key_str"', [Mode.ROOT], State.EXPECT_COLON),
    ('{"key_str":', [Mode.ROOT], State.EXPECT_VALUE),
    ('{"key_str": ', [Mode.ROOT], State.EXPECT_VALUE),
    ('{"key_str": \t\n', [Mode.ROOT], State.EXPECT_VALUE),
    ('{"key_str": "', [Mode.ROOT], State.PARSING_STRING),
    ('{"key_str": "val', [Mode.ROOT], State.PARSING_STRING),
    ('{"key_str": "val"', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val" \t\n', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val",', [Mode.ROOT], State.EXPECT_KEY_AFTER_COMMA),
    ('{"key_str": "val","key_int', [Mode.ROOT], State.PARSING_KEY),
    ('{"key_str": "val","key_int"', [Mode.ROOT], State.EXPECT_COLON),
    ('{"key_str": "val","key_int":', [Mode.ROOT], State.EXPECT_VALUE),
    ('{"key_str": "val","key_int": \t\n', [Mode.ROOT], State.EXPECT_VALUE),
    ('{"key_str": "val","key_int":4', [Mode.ROOT], State.PARSING_INTEGER),
    ('{"key_str": "val","key_int":42', [Mode.ROOT], State.PARSING_INTEGER),
    ('{"key_str": "val","key_int":42,', [Mode.ROOT], State.EXPECT_KEY_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"', [Mode.ROOT], State.PARSING_KEY),
    ('{"key_str": "val","key_int":42,"key_float"', [Mode.ROOT], State.EXPECT_COLON),
    ('{"key_str": "val","key_int":42,"key_float":', [Mode.ROOT], State.EXPECT_VALUE),
    ('{"key_str": "val","key_int":42,"key_float":', [Mode.ROOT], State.EXPECT_VALUE),
    ('{"key_str": "val","key_int":42,"key_float":3.14', [Mode.ROOT], State.PARSING_FLOAT),
    ('{"key_str": "val","key_int":42,"key_float":3.14,', [Mode.ROOT], State.EXPECT_KEY_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":', [Mode.ROOT], State.EXPECT_VALUE),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":t', [Mode.ROOT], State.PARSING_BOOLEAN),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true', [Mode.ROOT], State.PARSING_BOOLEAN),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":n', [Mode.ROOT], State.PARSING_NULL),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,', [Mode.ROOT], State.EXPECT_KEY_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,', [Mode.ROOT], State.EXPECT_KEY_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream', [Mode.ROOT], State.PARSING_KEY),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream', [Mode.ROOT], State.PARSING_STRING),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum', [Mode.ROOT], State.PARSING_KEY),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"val', [Mode.ROOT], State.PARSING_STRING),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1"', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":', [Mode.ROOT], State.EXPECT_VALUE),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{', [Mode.ROOT, Mode.OBJECT], State.PARSING_OBJECT),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"', [Mode.ROOT, Mode.OBJECT], State.PARSING_OBJECT),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str"', [Mode.ROOT, Mode.OBJECT], State.PARSING_OBJECT),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"', [Mode.ROOT, Mode.OBJECT], State.PARSING_OBJECT),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"}', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},', [Mode.ROOT], State.EXPECT_KEY_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":', [Mode.ROOT], State.EXPECT_VALUE),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":[', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["', [Mode.ROOT, Mode.ARRAY], State.PARSING_STRING),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1" \t\n', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1",', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1", \t\n', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2",', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3', [Mode.ROOT, Mode.ARRAY], State.PARSING_STRING),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],', [Mode.ROOT], State.EXPECT_KEY_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42', [Mode.ROOT, Mode.ARRAY], State.PARSING_INTEGER),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3', [Mode.ROOT, Mode.ARRAY], State.PARSING_FLOAT),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14', [Mode.ROOT, Mode.ARRAY], State.PARSING_FLOAT),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true', [Mode.ROOT, Mode.ARRAY], State.PARSING_BOOLEAN),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false', [Mode.ROOT, Mode.ARRAY], State.PARSING_BOOLEAN),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,nul', [Mode.ROOT, Mode.ARRAY], State.PARSING_NULL),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":[', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["val', [Mode.ROOT, Mode.ARRAY], State.PARSING_STRING),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1"', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{', [Mode.ROOT, Mode.ARRAY, Mode.OBJECT], State.PARSING_OBJECT),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_s', [Mode.ROOT, Mode.ARRAY, Mode.OBJECT], State.PARSING_OBJECT),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"}', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},', [Mode.ROOT, Mode.ARRAY], State.EXPECT_VALUE_AFTER_COMMA),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nes', [Mode.ROOT, Mode.ARRAY, Mode.OBJECT], State.PARSING_OBJECT),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}', [Mode.ROOT, Mode.ARRAY], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}]', [Mode.ROOT], State.EXPECT_COMMA_OR_EOC),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,"key_none":null,"key_stream":"stream","key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1","val2","val3"],"arr_int":[42,43],"arr_float":[3.14,31.4],"arr_bool":[true,false,true],"arr_none":[null,null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}]}', [], State.END),
]
# fmt: on


@pytest.mark.parametrize(
    "stream,expected_stack,expected_state",
    parse_correct_stream__params,
)
@pytest.mark.anyio
async def test_json_demux__parse_correct_stream__assert_state(
    stream: str, expected_stack: List[Mode], expected_state: State
):
    class SObject(JMux):
        class SNested(JMux):
            key_str: AwaitableValue[str]

        class SEnum(Enum):
            VALUE1 = "value1"
            VALUE2 = "value2"

        key_str: AwaitableValue[str]
        key_int: AwaitableValue[int]
        key_float: AwaitableValue[float]
        key_bool: AwaitableValue[bool]
        key_none: AwaitableValue[NoneType]
        key_stream: StreamableValues[str]
        key_enum: AwaitableValue[SEnum]
        key_nested: AwaitableValue[SNested]

        arr_str: StreamableValues[str]
        arr_int: StreamableValues[int]
        arr_float: StreamableValues[float]
        arr_bool: StreamableValues[bool]
        arr_none: StreamableValues[NoneType]
        arr_enum: StreamableValues[SEnum]
        arr_nested: StreamableValues[SNested]

    s_object = SObject()

    for ch in stream:
        await s_object.feed_char(ch)

    assert s_object._pda.state == expected_state
    assert s_object._pda._stack == expected_stack


# fmt: off
parse_incorrect_stream__params = [
    ("b", UnexpectedCharacterError),
    ("\n", None),
    (" ", None),
    ("\t", None),
    ("{", None),
    ("{p", UnexpectedCharacterError),
    ('{"', None),
    ('{""', EmptyKeyError),
    ('{"no_actual_key"', MissingAttributeError),
    ('{"key_str"', None),
    ('{"key_str": ""', None),
    ('{"key_str": "" ', None),
    ('{"key_str": "val",}', UnexpectedCharacterError), # Trailing comma in object
    ('{"key_str": "val","key_int":4p', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":4t', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":420', None),
    ('{"key_str": "val","key_int":420 ', None),
    ('{"key_str": "val","key_int":-420', None),
    ('{"key_str": "val","key_int":-4.20', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":1e+', None),
    ('{"key_str": "val","key_int":42,"key_float":0', None),
    ('{"key_str": "val","key_int":42,"key_float":p', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":1e+,', ParsePrimitiveError),
    ('{"key_str": "val","key_int":42,"key_float":-3.14e10,', None),
    ('{"key_str": "val","key_int":42,"key_float":-2.5E3,', None),
    ('{"key_str": "val","key_int":42,"key_float":1E+10,', None),
    ('{"key_str": "val","key_int":42,"key_float":NaN', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":Infinity', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":-', None),
    ('{"key_str": "val","key_int":42,"key_float":- ', ParsePrimitiveError),
    ('{"key_str": "val","key_int":42,"key_float":+', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":-1', None),
    ('{"key_str": "val","key_int":42,"key_float":-1 ', None),
    ('{"key_str": "val","key_int":42,"key_float":--1', None),
    ('{"key_str": "val","key_int":42,"key_float":--1,', ParsePrimitiveError),
    ('{"key_str": "val","key_int":42,"key_float":.', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":1.', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":t', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":t ', ParsePrimitiveError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":T', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":trub', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":tf', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":trueee', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true ', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":f', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":F', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":ft', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":falsb', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":n', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":n ', ParsePrimitiveError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":nope', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":nulll', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null ', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":,', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"val', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"foo', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1"', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"foobar"', ParsePrimitiveError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":p', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":n', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":4', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{p', UnexpectedCharacterError), # Means all recursive calls throw errors as expected
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"} ', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":{', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":p', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":[[', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":[]', None), # Allow empty arrays
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1",]', UnexpectedCharacterError), # Trailing comma in array
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":[nu', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1",}', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":4,', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[4.', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42,', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42,[', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[-42,', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42,+43]', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42,-43]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":3', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":{', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":"', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3k', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[0', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14,314]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3,1,4]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14,31.4]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":"', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":t', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":r', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true,false,true]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":n', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":f', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum"', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":[', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["val', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["foo', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1"', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["foobar"', ParsePrimitiveError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[]', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[3', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[p', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{p', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":3', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}]}', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}]}\n', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}]} ', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}]}\t', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_none":null,"key_enum":"value1","key_nested":{"key_str":"nested"},"arr_str":["val1"],"arr_int":[42],"arr_float":[3.14],"arr_bool":[true],"arr_none":[null],"arr_enum":["value1","value2"],"arr_nested":[{"key_str":"nested1"},{"key_str":"nested2"}]}}', ObjectAlreadyClosedError),
]
# fmt: on


@pytest.mark.parametrize(
    "stream,MaybeExpectedError",
    parse_incorrect_stream__params,
)
@pytest.mark.anyio
async def test_json_demux__parse_stream__assert_error(
    stream: str, MaybeExpectedError: Type[Exception] | None
):
    class SObject(JMux):
        class SNested(JMux):
            key_str: AwaitableValue[str]

        class SEnum(Enum):
            VALUE1 = "value1"
            VALUE2 = "value2"

        key_str: AwaitableValue[str]
        key_int: AwaitableValue[int]
        key_float: AwaitableValue[float]
        key_bool: AwaitableValue[bool]
        key_none: AwaitableValue[NoneType]
        key_stream: StreamableValues[str]
        key_enum: AwaitableValue[SEnum]
        key_nested: AwaitableValue[SNested]

        arr_str: StreamableValues[str]
        arr_int: StreamableValues[int]
        arr_float: StreamableValues[float]
        arr_bool: StreamableValues[bool]
        arr_none: StreamableValues[NoneType]
        arr_enum: StreamableValues[SEnum]
        arr_nested: StreamableValues[SNested]

    s_object = SObject()

    if MaybeExpectedError:
        with pytest.raises(MaybeExpectedError):
            for ch in stream:
                await s_object.feed_char(ch)
    else:
        for ch in stream:
            await s_object.feed_char(ch)


# fmt: off
parse_incorrect_stream_with_optionals__params = [
    ("b", UnexpectedCharacterError),
    ("{", None),
    ("{p", UnexpectedCharacterError),
    ('{"', None),
    ('{""', EmptyKeyError),
    ('{"no_actual_key"', MissingAttributeError),
    ('{"key_str"', None),
    ('{"key_str": ""', None),
    ('{"key_str": n', None),
    ('{"key_str": t', UnexpectedCharacterError),
    ('{"key_str": null', None),
    ('{"key_str": "val","key_int":4p', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":n', None),
    ('{"key_str": "val","key_int":r', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":null', None),
    ('{"key_str": "val","key_int":null,', None),
    ('{"key_str": "val","key_int":420', None),
    ('{"key_str": "val","key_int":-420', None),
    ('{"key_str": "val","key_int":42,"key_float":0', None),
    ('{"key_str": "val","key_int":42,"key_float":n', None),
    ('{"key_str": "val","key_int":42,"key_float":l', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":null', None),
    ('{"key_str": "val","key_int":42,"key_float":null,', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":t', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":r', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":true,', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":n', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":null', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":null,', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":t', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":n', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":null,', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1"', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":{', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":n', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":k', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":null', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":null ', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":null,', None),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":{p', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":{n', UnexpectedCharacterError),
    ('{"key_str": "val","key_int":42,"key_float":3.14,"key_bool":false,"key_enum":"value1","key_nested":{"key_str":"nested"},', None),
]
# fmt: on


@pytest.mark.parametrize(
    "stream,MaybeExpectedError",
    parse_incorrect_stream_with_optionals__params,
)
@pytest.mark.anyio
async def test_json_demux__parse_stream_with_optionals__assert_error(
    stream: str, MaybeExpectedError: Type[Exception] | None
):
    class SObject(JMux):
        class SNested(JMux):
            key_str: AwaitableValue[str]

        class SEnum(Enum):
            VALUE1 = "value1"
            VALUE2 = "value2"

        key_str: AwaitableValue[str | NoneType]
        key_int: AwaitableValue[int | NoneType]
        key_float: AwaitableValue[float | NoneType]
        key_bool: AwaitableValue[bool | NoneType]
        key_enum: AwaitableValue[SEnum | NoneType]
        key_nested: AwaitableValue[SNested | NoneType]

    s_object = SObject()

    if MaybeExpectedError:
        with pytest.raises(MaybeExpectedError):
            for ch in stream:
                await s_object.feed_char(ch)
    else:
        for ch in stream:
            await s_object.feed_char(ch)


# fmt: off
parse_correct_stream__double_nested__params = [
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"', None),
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"}', None),
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"}}', None),
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"}}}', None),
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"}, ', None),
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"}, "key_str":', None), # Simplified version of the trigger encountered in main repo
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"}, "key_str": t', UnexpectedCharacterError),
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"}, "key_str": "val', None),
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"}, "key_str": "val"}', None),
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"}, "key_str": "val"}}', None),
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"}, "key_str": null}}', None),
    ('{"key_first_nested": {"key_second_nested": {"key_str": "val"}, "key_str": null\n}}', None),
    ('{"key_first_nested": null}', None),
    ('{"key_first_nested": null\n}', None),
]
# fmt: on
@pytest.mark.parametrize(
    "stream,MaybeExpectedError",
    parse_correct_stream__double_nested__params,
)
@pytest.mark.anyio
async def test_json_demux__parse_stream__double_nested(
    stream: str, MaybeExpectedError: Type[Exception] | None
):
    class SObject(JMux):
        class SFirstNested(JMux):
            class SSecondNested(JMux):
                key_str: AwaitableValue[str]

            key_second_nested: AwaitableValue[SSecondNested | NoneType]
            key_str: AwaitableValue[str | NoneType]

        key_first_nested: AwaitableValue[SFirstNested | NoneType]

    s_object = SObject()

    if MaybeExpectedError:
        with pytest.raises(MaybeExpectedError):
            for ch in stream:
                await s_object.feed_char(ch)
    else:
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__type_mismatch__string_when_int_expected():
    class SObject(JMux):
        key_int: AwaitableValue[int]

    s_object = SObject()
    stream = '{"key_int": "42"}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__type_mismatch__int_when_string_expected():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": 42}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__type_mismatch__array_when_object_expected():
    class SObject(JMux):
        class SNested(JMux):
            key_str: AwaitableValue[str]

        key_nested: AwaitableValue[SNested]

    s_object = SObject()
    stream = '{"key_nested": []}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__type_mismatch__object_when_array_expected():
    class SObject(JMux):
        arr_str: StreamableValues[str]

    s_object = SObject()
    stream = '{"arr_str": {}}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__type_mismatch__boolean_when_number_expected():
    class SObject(JMux):
        key_int: AwaitableValue[int]

    s_object = SObject()
    stream = '{"key_int": true}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__type_mismatch__null_when_non_optional_expected():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": null}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__type_mismatch__float_when_int_expected():
    class SObject(JMux):
        key_int: AwaitableValue[int]

    s_object = SObject()
    stream = '{"key_int": 3.14}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__type_mismatch__string_when_bool_expected():
    class SObject(JMux):
        key_bool: AwaitableValue[bool]

    s_object = SObject()
    stream = '{"key_bool": "true"}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__type_mismatch__number_when_bool_expected():
    class SObject(JMux):
        key_bool: AwaitableValue[bool]

    s_object = SObject()
    stream = '{"key_bool": 1}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__type_mismatch__string_in_int_array():
    class SObject(JMux):
        arr_int: StreamableValues[int]

    s_object = SObject()
    stream = '{"arr_int": [1, "2", 3]}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__type_mismatch__int_in_string_array():
    class SObject(JMux):
        arr_str: StreamableValues[str]

    s_object = SObject()
    stream = '{"arr_str": ["a", 2, "c"]}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__feed_char_with_multi_char_string_raises():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    with pytest.raises(UnexpectedCharacterError):
        await s_object.feed_char("ab")


@pytest.mark.anyio
async def test_json_demux__feed_char_after_close_raises():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "val"}'
    for ch in stream:
        await s_object.feed_char(ch)

    with pytest.raises(ObjectAlreadyClosedError):
        await s_object.feed_char("{")


@pytest.mark.anyio
async def test_json_demux__feed_char_empty_string():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    with pytest.raises(UnexpectedCharacterError):
        await s_object.feed_char("")


@pytest.mark.anyio
async def test_json_demux__trailing_comma_in_object():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "val",}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__trailing_comma_in_array():
    class SObject(JMux):
        arr_str: StreamableValues[str]

    s_object = SObject()
    stream = '{"arr_str": ["a", "b",]}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__leading_zero_in_integer():
    class SObject(JMux):
        key_int: AwaitableValue[int]

    s_object = SObject()
    stream = '{"key_int": 01}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__very_large_integer():
    class SObject(JMux):
        key_int: AwaitableValue[int]

    s_object = SObject()
    stream = '{"key_int": 99999999999999999999999999999999}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__very_small_integer():
    class SObject(JMux):
        key_int: AwaitableValue[int]

    s_object = SObject()
    stream = '{"key_int": -99999999999999999999999999999999}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__very_large_float():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": 1.7976931348623157e+308}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__very_small_float():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": 5e-324}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__negative_zero():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": -0}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__deeply_nested_objects():
    class Level3(JMux):
        value: AwaitableValue[str]

    class Level2(JMux):
        level3: AwaitableValue[Level3]

    class Level1(JMux):
        level2: AwaitableValue[Level2]

    class SObject(JMux):
        level1: AwaitableValue[Level1]

    s_object = SObject()
    stream = '{"level1": {"level2": {"level3": {"value": "deep"}}}}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__control_characters_in_string():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "hello\\nworld\\ttab"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__unicode_in_key():
    class SObject(JMux):
        日本語: AwaitableValue[str]

    s_object = SObject()
    stream = '{"日本語": "value"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__unicode_in_value():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "日本語"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__emoji_in_value():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "🎉🚀"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__empty_object():
    class SObject(JMux):
        pass

    s_object = SObject()
    stream = "{}"

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__empty_array():
    class SObject(JMux):
        arr_str: StreamableValues[str]

    s_object = SObject()
    stream = '{"arr_str": []}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__whitespace_only_before_object():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '   \n\t  {"key_str": "val"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__whitespace_within_object():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{   "key_str"  :   "val"   }'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__nested_arrays_not_allowed():
    class SObject(JMux):
        arr: StreamableValues[str]

    s_object = SObject()
    stream = '{"arr": [["a"]]}'

    with pytest.raises(UnexpectedCharacterError):
        for ch in stream:
            await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__double_quote_escaped_in_string():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "hello \\"world\\""}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__backslash_escaped_in_string():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "C:\\\\Users\\\\Name"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__integer_zero():
    class SObject(JMux):
        key_int: AwaitableValue[int]

    s_object = SObject()
    stream = '{"key_int": 0}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__float_zero():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": 0.0}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__scientific_notation_negative_exponent():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": 1.5e-10}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__scientific_notation_positive_exponent():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": 1.5e+10}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__scientific_notation_capital_e():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": 1.5E10}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__multiple_keys_same_type():
    class SObject(JMux):
        key1: AwaitableValue[str]
        key2: AwaitableValue[str]
        key3: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key1": "val1", "key2": "val2", "key3": "val3"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__mixed_types():
    class SObject(JMux):
        key_str: AwaitableValue[str]
        key_int: AwaitableValue[int]
        key_float: AwaitableValue[float]
        key_bool: AwaitableValue[bool]

    s_object = SObject()
    stream = '{"key_str": "val", "key_int": 42, "key_float": 3.14, "key_bool": true}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__feed_chunks_simple():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    await s_object.feed_chunks('{"key_str": "val"}')


@pytest.mark.anyio
async def test_json_demux__feed_chunks_multi_call():
    class SObject(JMux):
        key_str: AwaitableValue[str]
        key_int: AwaitableValue[int]

    s_object = SObject()
    await s_object.feed_chunks('{"key_str": "val",')
    await s_object.feed_chunks('"key_int": 42}')


@pytest.mark.anyio
async def test_json_demux__array_with_many_elements():
    class SObject(JMux):
        arr_int: StreamableValues[int]

    s_object = SObject()
    arr_content = ",".join(str(i) for i in range(100))
    stream = f'{{"arr_int": [{arr_content}]}}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_json_demux__long_string_value():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    long_value = "a" * 10000
    stream = f'{{"key_str": "{long_value}"}}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__10_levels_deep():
    class L10(JMux):
        val: AwaitableValue[str]

    class L9(JMux):
        l10: AwaitableValue[L10]

    class L8(JMux):
        l9: AwaitableValue[L9]

    class L7(JMux):
        l8: AwaitableValue[L8]

    class L6(JMux):
        l7: AwaitableValue[L7]

    class L5(JMux):
        l6: AwaitableValue[L6]

    class L4(JMux):
        l5: AwaitableValue[L5]

    class L3(JMux):
        l4: AwaitableValue[L4]

    class L2(JMux):
        l3: AwaitableValue[L3]

    class L1(JMux):
        l2: AwaitableValue[L2]

    class SObject(JMux):
        l1: AwaitableValue[L1]

    s_object = SObject()
    stream = '{"l1":{"l2":{"l3":{"l4":{"l5":{"l6":{"l7":{"l8":{"l9":{"l10":{"val":"deep"}}}}}}}}}}}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__1000_char_string():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    long_value = "x" * 1000
    stream = f'{{"key_str": "{long_value}"}}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__10000_char_string():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    long_value = "y" * 10000
    stream = f'{{"key_str": "{long_value}"}}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__long_key_name():
    long_key = "k" * 500

    SObject = type(
        "SObject", (JMux,), {"__annotations__": {long_key: AwaitableValue[str]}}
    )

    s_object = SObject()
    stream = f'{{"{long_key}": "value"}}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__negative_with_exponent():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": -1.5e10}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__very_large_exponent():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": 1e308}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__very_small_exponent():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": 1e-308}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__leading_zero_float():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": 0.123}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__trailing_zeros():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": 1.10000}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__negative_zero_float():
    class SObject(JMux):
        key_float: AwaitableValue[float]

    s_object = SObject()
    stream = '{"key_float": -0.0}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__integer_max_safe():
    class SObject(JMux):
        key_int: AwaitableValue[int]

    s_object = SObject()
    stream = '{"key_int": 9007199254740991}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__integer_beyond_js_safe():
    class SObject(JMux):
        key_int: AwaitableValue[int]

    s_object = SObject()
    stream = '{"key_int": 9007199254740992}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__very_long_integer():
    class SObject(JMux):
        key_int: AwaitableValue[int]

    s_object = SObject()
    stream = '{"key_int": 123456789012345678901234567890}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__carriage_return():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{\r"key_str": "val"\r}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__mixed_whitespace():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{ \t\n\r "key_str" \t\n\r : \t\n\r "val" \t\n\r }'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__whitespace_before_array_close():
    class SObject(JMux):
        arr: StreamableValues[int]

    s_object = SObject()
    stream = '{"arr": [1, 2, 3   ]}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__whitespace_before_object_close():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "val"   }'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__newline_after_value_before_close():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "val"\n}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__array_with_whitespace_between_elements():
    class SObject(JMux):
        arr: StreamableValues[int]

    s_object = SObject()
    stream = '{"arr": [1 , 2 , 3]}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__single_element_array():
    class SObject(JMux):
        arr: StreamableValues[str]

    s_object = SObject()
    stream = '{"arr": ["only"]}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__string_with_backslash_u():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "\\u0041"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__string_with_unicode_escape_chinese():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "\\u4e2d\\u6587"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__string_with_null_character():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "\\u0000"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__string_with_solidus():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "a\\/b"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__string_only_escapes():
    class SObject(JMux):
        key_str: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key_str": "\\n\\t\\r\\b\\f"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__object_with_single_field():
    class SObject(JMux):
        key: AwaitableValue[str]

    s_object = SObject()
    stream = '{"key": "value"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__object_key_with_spaces():
    class SObject(JMux):
        pass

    SObject.__annotations__["key with spaces"] = AwaitableValue[str]

    s_object = SObject()
    stream = '{"key with spaces": "value"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__object_key_with_unicode():
    class SObject(JMux):
        キー: AwaitableValue[str]

    s_object = SObject()
    stream = '{"キー": "値"}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__boolean_false_in_array():
    class SObject(JMux):
        arr: StreamableValues[bool]

    s_object = SObject()
    stream = '{"arr": [false]}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__all_null_array():
    class SObject(JMux):
        arr: StreamableValues[NoneType]

    s_object = SObject()
    stream = '{"arr": [null, null, null]}'

    for ch in stream:
        await s_object.feed_char(ch)


@pytest.mark.anyio
async def test_demux_parse__mixed_booleans():
    class SObject(JMux):
        arr: StreamableValues[bool]

    s_object = SObject()
    stream = '{"arr": [true, false, true, false]}'

    for ch in stream:
        await s_object.feed_char(ch)



from enum import Enum
from typing import Sequence


class MissingAttributeError(Exception):
    def __init__(self, object_name: str, attribute: str) -> None:
        super().__init__(f"'{object_name}' is missing required attribute '{attribute}'")


class UnexpectedAttributeTypeError(Exception):
    def __init__(self, object_name: str, attribute: str, expected_type: str) -> None:
        super().__init__(
            f"'{object_name}' has attribute '{attribute}' with unexpected type. "
            f"Attribute must conform to {expected_type}."
        )


class EmptyKeyError(Exception):
    def __init__(self, message: str | None = None) -> None:
        super().__init__("Key cannot be empty" + (f": {message}" if message else ""))
        self.message = message


class TypeEmitError(Exception):
    def __init__(self, expected_type: str, actual_type: str) -> None:
        super().__init__(
            f"Cannot emit to current sink. Type mismatch: expected {expected_type}, "
            f"got {actual_type}"
        )
        self.expected_type = expected_type
        self.actual_type = actual_type


class NoCurrentSinkError(Exception):
    def __init__(self, message: str | None = None) -> None:
        super().__init__(
            "No current sink available" + (f": {message}" if message else "")
        )


class ParsePrimitiveError(Exception):
    def __init__(self, message: str | None = None) -> None:
        super().__init__(
            "Failed to parse primitive value" + (f": {message}" if message else "")
        )


class NothingEmittedError(Exception):
    def __init__(self, message: str | None = None) -> None:
        super().__init__(
            "Nothing was emitted to the sink" + (f": {message}" if message else "")
        )
        self.message = message


class SinkClosedError(Exception):
    def __init__(self, message: str | None = None) -> None:
        super().__init__(
            "Sink is closed and cannot accept new items"
            + (f": {message}" if message else "")
        )
        self.message = message


class NotAllObjectPropertiesSetError(Exception):
    def __init__(self, message: str | None = None) -> None:
        super().__init__(
            "Not all properties were set before closing the sink respective stream"
            + (f": {message}" if message else "")
        )


class ObjectAlreadyClosedError(Exception):
    def __init__(self, object_name: str, message: str | None = None) -> None:
        super().__init__(
            f"Object '{object_name}' is already closed"
            + (f": {message}" if message else "")
        )


class ObjectMissmatchedError(Exception):
    def __init__(
        self,
        jmux_model: str,
        pydantic_model: str,
        attribute: str,
        message: str | None = None,
    ) -> None:
        super().__init__(
            f"JMux object '{jmux_model}' and '{pydantic_model}' are missmatched on "
            f"attribute '{attribute}'" + (f": {message}" if message else "")
        )


class ForbiddenTypeHintsError(Exception):
    def __init__(self, message: str | None = None) -> None:
        super().__init__(
            "Forbidden type hints used in object definition"
            + (f": {message}" if message else "")
        )


class UnexpectedCharacterError(Exception):
    def __init__(
        self,
        character: str,
        pda_stack: Sequence[Enum] | str,
        pda_state: Enum | str,
        message: str | None,
    ) -> None:
        pda_stack_str = [
            item.value if isinstance(item, Enum) else item for item in pda_stack
        ]
        pda_state_str = pda_state.value if isinstance(pda_state, Enum) else pda_state
        super().__init__(
            f"Received unexpected character '{character}' in state '{pda_state_str}' "
            f"with stack {pda_stack_str}" + (f": {message}" if message else "")
        )


class StreamParseError(Exception):
    def __init__(self, message: str | None = None) -> None:
        super().__init__("Failed to parse stream" + (f": {message}" if message else ""))
        self.message = message


class UnexpectedStateError(Exception):
    def __init__(
        self,
        pda_stack: Sequence[Enum] | str,
        pda_state: Enum | str,
        message: str | None = None,
    ) -> None:
        pda_stack_str = [
            item.value if isinstance(item, Enum) else item for item in pda_stack
        ]
        pda_state_str = pda_state.value if isinstance(pda_state, Enum) else pda_state
        super().__init__(
            f"Unexpected state '{pda_state_str}' with stack {pda_stack_str}"
            + (f": {message}" if message else "")
        )



from typing import Protocol


class IDecoder(Protocol):
    def push(self, ch: str) -> str | None: ...

    def is_terminating_quote(self, ch: str) -> bool: ...

    def reset(self) -> None: ...

    @property
    def buffer(self) -> str: ...


class StringEscapeDecoder:
    r"""
    Decoder for strings with escape sequences, such as JSON strings.
    Handles escape sequences like \", \\, \/, \b, \f, \n, \r, \t, and unicode escapes.
    """

    escape_map = {
        '"': '"',
        "\\": "\\",
        "/": "/",
        "b": "\b",
        "f": "\f",
        "n": "\n",
        "r": "\r",
        "t": "\t",
    }

    def __init__(self):
        self._buffer = ""
        self._string_escape = False
        self._is_parsing_unicode = False
        self._unicode_buffer = ""

    def push(self, ch: str) -> str | None:
        if self._is_parsing_unicode:
            self._unicode_buffer += ch
            if len(self._unicode_buffer) == 4:
                code_point = int(self._unicode_buffer, 16)
                decoded_char = chr(code_point)
                self._buffer += decoded_char
                self._is_parsing_unicode = False
                self._unicode_buffer = ""
                return decoded_char
            return None

        if self._string_escape:
            self._string_escape = False
            if ch == "u":
                self._is_parsing_unicode = True
                self._unicode_buffer = ""
                return None
            escaped_char = self.escape_map.get(ch, ch)
            self._buffer += escaped_char
            return escaped_char

        if ch == "\\":
            self._string_escape = True
            return None
        else:
            self._buffer += ch
            return ch

    def is_terminating_quote(self, ch: str) -> bool:
        if self._string_escape or self._is_parsing_unicode:
            return False
        if ch == '"':
            return True
        return False

    def reset(self) -> None:
        self._buffer = ""
        self._string_escape = False
        self._is_parsing_unicode = False
        self._unicode_buffer = ""

    @property
    def buffer(self) -> str:
        return self._buffer



from enum import Enum

import pytest
from pydantic import BaseModel

from jmux.awaitable import AwaitableValue, StreamableValues
from jmux.demux import JMux
from jmux.error import (
    ForbiddenTypeHintsError,
    MissingAttributeError,
    ObjectMissmatchedError,
)


class SEnum(Enum):
    VALUE1 = "value1"
    VALUE2 = "value2"


class CorrectJMux_1(JMux):
    class NestedJMux(JMux):
        nested_key: AwaitableValue[str]

    key_str: AwaitableValue[str]
    key_int: AwaitableValue[int]
    key_float: AwaitableValue[float]
    key_bool: AwaitableValue[bool]
    key_none: AwaitableValue[None]
    key_enum: AwaitableValue[SEnum]
    key_nested: AwaitableValue[NestedJMux]

    key_stream: StreamableValues[str]

    arr_str: StreamableValues[str]
    arr_int: StreamableValues[int]
    arr_float: StreamableValues[float]
    arr_bool: StreamableValues[bool]
    arr_none: StreamableValues[None]
    arr_enum: StreamableValues[SEnum]
    arr_nested: StreamableValues[NestedJMux]


class CorrectPydantic_1(BaseModel):
    class NestedPydantic(BaseModel):
        nested_key: str

    key_str: str
    key_int: int
    key_float: float
    key_bool: bool
    key_none: None
    key_stream: str
    key_enum: SEnum
    key_nested: NestedPydantic

    arr_str: list[str]
    arr_int: list[int]
    arr_float: list[float]
    arr_bool: list[bool]
    arr_none: list[None]
    arr_enum: list[SEnum]
    arr_nested: list[NestedPydantic]


class CorrectJMux_2(JMux):
    key_str: AwaitableValue[str | None]
    key_int: AwaitableValue[int | None]
    key_float: AwaitableValue[float | None]
    key_bool: AwaitableValue[bool | None]


class CorrectPydantic_2(BaseModel):
    key_str: str | None
    key_int: int | None
    key_float: float | None
    key_bool: bool | None


class CorrectJMux_3(JMux):
    arr_str: StreamableValues[str]


class CorrectPydantic_3(BaseModel):
    arr_str: list[str] | None


class IncorrectJMux_1(JMux):
    key_str: AwaitableValue[str]


class IncorrectPydantic_1(BaseModel):
    key_str: str | None


class IncorrectJMux_2(JMux):
    key_str: AwaitableValue[str | None]


class IncorrectPydantic_2(BaseModel):
    key_str: str


class IncorrectJMux_3(JMux):
    class NestedJMux(JMux):
        nested_key: AwaitableValue[str | None]

    key_nested: AwaitableValue[NestedJMux]


class IncorrectPydantic_3(BaseModel):
    class NestedPydantic(BaseModel):
        nested_key: str

    key_nested: NestedPydantic


@pytest.mark.parametrize(
    "TargetJMux,TargetPydantic,MaybeExpectedError",
    [
        (CorrectJMux_1, CorrectPydantic_1, None),
        (CorrectJMux_2, CorrectPydantic_2, None),
        (CorrectJMux_3, CorrectPydantic_3, None),
        (IncorrectJMux_1, IncorrectPydantic_1, ObjectMissmatchedError),
        (IncorrectJMux_2, IncorrectPydantic_2, ObjectMissmatchedError),
        (IncorrectJMux_3, IncorrectPydantic_3, ObjectMissmatchedError),
    ],
)
@pytest.mark.anyio
async def test_json_demux__validate_pydantic(
    TargetJMux: type[JMux],
    TargetPydantic: type[BaseModel],
    MaybeExpectedError: type[Exception] | None,
):
    if MaybeExpectedError:
        with pytest.raises(MaybeExpectedError):
            TargetJMux.assert_conforms_to(TargetPydantic)
    else:
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__missing_attribute_error():
    class TargetJMux(JMux):
        required_field: AwaitableValue[str]

    class TargetPydantic(BaseModel):
        pass

    with pytest.raises(MissingAttributeError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__missing_attribute_error_with_optional_passes():
    class TargetJMux(JMux):
        optional_field: AwaitableValue[str | None]

    class TargetPydantic(BaseModel):
        pass

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__nested_mismatch():
    class NestedJMux(JMux):
        field: AwaitableValue[int]

    class TargetJMux(JMux):
        nested: AwaitableValue[NestedJMux]

    class NestedPydantic(BaseModel):
        field: str

    class TargetPydantic(BaseModel):
        nested: NestedPydantic

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__streamable_values_type_mismatch():
    class TargetJMux(JMux):
        arr: StreamableValues[int]

    class TargetPydantic(BaseModel):
        arr: list[str]

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__streamable_values_string_type():
    class TargetJMux(JMux):
        stream: StreamableValues[str]

    class TargetPydantic(BaseModel):
        stream: str

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__streamable_values_string_type_mismatch():
    class TargetJMux(JMux):
        stream: StreamableValues[str]

    class TargetPydantic(BaseModel):
        stream: int

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__enum_matching():
    class TargetEnum(Enum):
        A = "a"
        B = "b"

    class TargetJMux(JMux):
        field: AwaitableValue[TargetEnum]

    class TargetPydantic(BaseModel):
        field: TargetEnum

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__enum_optional():
    class TargetEnum(Enum):
        A = "a"

    class TargetJMux(JMux):
        field: AwaitableValue[TargetEnum | None]

    class TargetPydantic(BaseModel):
        field: TargetEnum | None

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__streamable_enum_list():
    class TargetEnum(Enum):
        A = "a"
        B = "b"

    class TargetJMux(JMux):
        arr: StreamableValues[TargetEnum]

    class TargetPydantic(BaseModel):
        arr: list[TargetEnum]

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__nested_streamable():
    class NestedJMux(JMux):
        val: AwaitableValue[str]

    class TargetJMux(JMux):
        arr: StreamableValues[NestedJMux]

    class NestedPydantic(BaseModel):
        val: str

    class TargetPydantic(BaseModel):
        arr: list[NestedPydantic]

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__pydantic_dict_type_forbidden():
    class TargetJMux(JMux):
        field: AwaitableValue[str]

    class TargetPydantic(BaseModel):
        field: dict[str, str]

    with pytest.raises(ForbiddenTypeHintsError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__pydantic_tuple_type_forbidden():
    class TargetJMux(JMux):
        field: AwaitableValue[str]

    class TargetPydantic(BaseModel):
        field: tuple[str, int]

    with pytest.raises(ForbiddenTypeHintsError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__pydantic_set_type_forbidden():
    class TargetJMux(JMux):
        field: AwaitableValue[str]

    class TargetPydantic(BaseModel):
        field: set[str]

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__multiple_fields_all_match():
    class TargetJMux(JMux):
        field1: AwaitableValue[str]
        field2: AwaitableValue[int]
        field3: AwaitableValue[bool]

    class TargetPydantic(BaseModel):
        field1: str
        field2: int
        field3: bool

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__multiple_fields_one_mismatch():
    class TargetJMux(JMux):
        field1: AwaitableValue[str]
        field2: AwaitableValue[int]
        field3: AwaitableValue[bool]

    class TargetPydantic(BaseModel):
        field1: str
        field2: str
        field3: bool

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__extra_pydantic_field_allowed():
    class TargetJMux(JMux):
        field1: AwaitableValue[str]

    class TargetPydantic(BaseModel):
        field1: str
        field2: int

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__deep_nested_objects():
    class Level3JMux(JMux):
        val: AwaitableValue[str]

    class Level2JMux(JMux):
        level3: AwaitableValue[Level3JMux]

    class Level1JMux(JMux):
        level2: AwaitableValue[Level2JMux]

    class TargetJMux(JMux):
        level1: AwaitableValue[Level1JMux]

    class Level3Pydantic(BaseModel):
        val: str

    class Level2Pydantic(BaseModel):
        level3: Level3Pydantic

    class Level1Pydantic(BaseModel):
        level2: Level2Pydantic

    class TargetPydantic(BaseModel):
        level1: Level1Pydantic

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__deep_nested_mismatch_at_bottom():
    class Level2JMux(JMux):
        val: AwaitableValue[int]

    class Level1JMux(JMux):
        level2: AwaitableValue[Level2JMux]

    class TargetJMux(JMux):
        level1: AwaitableValue[Level1JMux]

    class Level2Pydantic(BaseModel):
        val: str

    class Level1Pydantic(BaseModel):
        level2: Level2Pydantic

    class TargetPydantic(BaseModel):
        level1: Level1Pydantic

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__mixed_streamable_and_awaitable():
    class TargetJMux(JMux):
        single_val: AwaitableValue[str]
        list_val: StreamableValues[int]
        stream_str: StreamableValues[str]

    class TargetPydantic(BaseModel):
        single_val: str
        list_val: list[int]
        stream_str: str

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__optional_list_in_pydantic():
    class TargetJMux(JMux):
        arr: StreamableValues[str]

    class TargetPydantic(BaseModel):
        arr: list[str] | None

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__all_primitive_types():
    class TargetJMux(JMux):
        str_val: AwaitableValue[str]
        int_val: AwaitableValue[int]
        float_val: AwaitableValue[float]
        bool_val: AwaitableValue[bool]
        none_val: AwaitableValue[None]

    class TargetPydantic(BaseModel):
        str_val: str
        int_val: int
        float_val: float
        bool_val: bool
        none_val: None

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_assert_conforms_to__all_optional_primitives():
    class TargetJMux(JMux):
        str_val: AwaitableValue[str | None]
        int_val: AwaitableValue[int | None]
        float_val: AwaitableValue[float | None]
        bool_val: AwaitableValue[bool | None]

    class TargetPydantic(BaseModel):
        str_val: str | None
        int_val: int | None
        float_val: float | None
        bool_val: bool | None

    TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__awaitable_int_vs_pydantic_str():
    class TargetJMux(JMux):
        field: AwaitableValue[int]

    class TargetPydantic(BaseModel):
        field: str

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__awaitable_str_vs_pydantic_int():
    class TargetJMux(JMux):
        field: AwaitableValue[str]

    class TargetPydantic(BaseModel):
        field: int

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__awaitable_bool_vs_pydantic_str():
    class TargetJMux(JMux):
        field: AwaitableValue[bool]

    class TargetPydantic(BaseModel):
        field: str

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__awaitable_float_vs_pydantic_int():
    class TargetJMux(JMux):
        field: AwaitableValue[float]

    class TargetPydantic(BaseModel):
        field: int

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__multiple_extra_non_optional_fields():
    class TargetJMux(JMux):
        field1: AwaitableValue[str]
        field2: AwaitableValue[int]
        field3: AwaitableValue[bool]

    class TargetPydantic(BaseModel):
        pass

    with pytest.raises(MissingAttributeError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__streamable_str_vs_list_int():
    class TargetJMux(JMux):
        arr: StreamableValues[str]

    class TargetPydantic(BaseModel):
        arr: list[int]

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__streamable_bool_vs_list_str():
    class TargetJMux(JMux):
        arr: StreamableValues[bool]

    class TargetPydantic(BaseModel):
        arr: list[str]

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__streamable_float_vs_list_int():
    class TargetJMux(JMux):
        arr: StreamableValues[float]

    class TargetPydantic(BaseModel):
        arr: list[int]

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__mismatched_enum_fails():
    class Enum1(Enum):
        A = "a"

    class Enum2(Enum):
        B = "b"

    class TargetJMux(JMux):
        field: AwaitableValue[Enum1]

    class TargetPydantic(BaseModel):
        field: Enum2

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__enum_vs_str_fails():
    class TargetEnum(Enum):
        A = "a"

    class TargetJMux(JMux):
        field: AwaitableValue[TargetEnum]

    class TargetPydantic(BaseModel):
        field: str

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__str_vs_enum_fails():
    class TargetEnum(Enum):
        A = "a"

    class TargetJMux(JMux):
        field: AwaitableValue[str]

    class TargetPydantic(BaseModel):
        field: TargetEnum

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__array_of_enum_mismatched():
    class Enum1(Enum):
        A = "a"

    class Enum2(Enum):
        B = "b"

    class TargetJMux(JMux):
        arr: StreamableValues[Enum1]

    class TargetPydantic(BaseModel):
        arr: list[Enum2]

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__nested_array_of_objects_mismatch():
    class NestedJMux(JMux):
        val: AwaitableValue[int]

    class TargetJMux(JMux):
        arr: StreamableValues[NestedJMux]

    class NestedPydantic(BaseModel):
        val: str

    class TargetPydantic(BaseModel):
        arr: list[NestedPydantic]

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)


def test_validation__streamable_int_vs_pydantic_str_fails():
    class TargetJMux(JMux):
        stream: StreamableValues[int]

    class TargetPydantic(BaseModel):
        stream: str

    with pytest.raises(ObjectMissmatchedError):
        TargetJMux.assert_conforms_to(TargetPydantic)
